-- --------------------------------------------------------------------
-- **                            Functions                           **
-- --------------------------------------------------------------------

local modes = {
	CHANGE = {"ChangeBar", 4, "EnhancedBars_ChangeBar_OnUpdate"},
	ABSORB = {"AbsorbBar", 5, nil},
	NECRO = {"NecroBar", 2, nil},
	OVERTIME = {"OverTimeBar", 1, nil},
};

local function cap(minValue, value, maxValue)
    return max(minValue, min(maxValue, value));
end

function EnhancedBars_CreateBar(parentBar, unit, mode)
    local info = modes[mode];
    assert(info);

    local bar = CreateFrame("StatusBar", parentBar:GetName()..info[1], parentBar);
    bar.mode = mode;
    bar.unit = unit;

    bar:SetFrameStrata(parentBar:GetFrameStrata());
    bar:SetFrameLevel(parentBar:GetFrameLevel() + info[2]);
    bar:SetStatusBarTexture("Interface\\TargetingFrame\\UI-StatusBar");

    if info[3] then bar:SetScript("OnUpdate", _G[info[3]]); end

    EnhancedBars_ResetBar(bar);

    return bar;
end

function EnhancedBars_ResetBar(bar)
    bar.guid = nil;
    bar.pause = 0;
    bar.active = false;
    bar.currentValue = nil;
    bar.maxValue = nil;
    bar.targetValue = nil;
    bar.speed = 0;
    bar:Hide();
end

local function finalizeCommon(bar, ratio, heightFactor, value)
    local parent = bar:GetParent();

    bar:SetWidth(parent:GetWidth());
    bar:SetHeight(parent:GetHeight() * heightFactor);
    bar:SetScale(parent:GetScale());
    bar:SetMinMaxValues(parent:GetMinMaxValues());
    bar:SetPoint("BOTTOMLEFT", parent, "BOTTOMLEFT", ratio * bar:GetWidth(), 0);
    bar:SetValue(value);
    bar:Show();
end

function EnhancedBars_FinalizeNormalBar(bar, shownValue)
    local delta = shownValue - bar.currentValue;
    local ratio;
    if delta >= 0 then
        ratio = bar.currentValue / bar.maxValue;
  else
        ratio = shownValue / bar.maxValue;
    end

    finalizeCommon(bar, ratio, 1, abs(delta));
end

function EnhancedBars_FinalizeMiniBar(bar, shownValue)
    local delta = shownValue - bar.currentValue;

    finalizeCommon(bar, 0, 0.5, abs(delta));
end

function EnhancedBars_ModifyChangeBar(bar, value, maxValue)
    assert(bar.mode == "CHANGE");
    maxValue = max(1, maxValue);
    value = cap(0, value, maxValue);

    if not bar.currentValue or not bar.maxValue then
        bar.pause = 0;
        bar.currentValue = value;
        bar.active = false;
    end

    bar.currentValue = min(bar.currentValue, maxValue); -- Never overflow the max
    bar.targetValue = value;
    bar.maxValue = maxValue;

    local delta = bar.targetValue - bar.currentValue;
    if delta ~= 0 then
        local mode = "DECREASE";
        if delta > 0 then mode = "INCREASE"; end

        local minSpeed, maxSpeed, adjustTime, pause, color = EnhancedBars_GetLayout(-1, mode);
        local speed = cap(minSpeed, math.abs(delta) / (adjustTime * maxValue), maxSpeed);

        if not bar.active then
            bar.pause = pause;
            bar.speed = speed;
            bar.active = true;
      else
            bar.speed = max(speed, bar.speed);
        end

        bar:SetStatusBarColor(color.r, color.g, color.b);
        EnhancedBars_ChangeBar_OnUpdate(bar, 0);
  else
        bar.active = false;
        bar:Hide();
    end
end

function EnhancedBars_ModifyPredictedBar(bar, predictedValue, value, maxValue)
    assert(bar.mode == "ABSORB" or bar.mode == "NECRO" or bar.mode == "OVERTIME");
    maxValue = max(1, maxValue);
    value = cap(0, value, maxValue);

    local finalizer;
    if predictedValue > value and bar.mode == "ABSORB" then
        finalizer = EnhancedBars_FinalizeMiniBar;
  else
        finalizer = EnhancedBars_FinalizeNormalBar;
        predictedValue = cap(0, predictedValue, maxValue);
    end

    bar.currentValue = value;
    bar.maxValue = maxValue;

    local delta = predictedValue - value;
    if delta ~= 0 then
        local mode;
        if bar.mode == "OVERTIME" then
            if delta > 0 then
                mode = "HOT";
          else
                mode = "DOT";
            end
      else
            mode = bar.mode;
        end

        local minSpeed, maxSpeed, adjustTime, pause, color = EnhancedBars_GetLayout(-1, mode);

        bar:SetStatusBarColor(color.r, color.g, color.b);
        finalizer(bar, predictedValue);
  else
        bar:Hide();
    end
end

function EnhancedBars_ChangeBar_OnUpdate(bar, elapsed)
    local delta = bar.targetValue - bar.currentValue;
    if delta ~= 0 then
        if bar.pause > 0 then
            bar.pause = max(0, bar.pause - elapsed);
      else
            local change = bar.speed * bar.maxValue * elapsed;
            if delta > 0 then
                bar.currentValue = min(bar.targetValue, bar.currentValue + change);
          else
                bar.currentValue = max(bar.targetValue, bar.currentValue - change);
            end
        end

        EnhancedBars_FinalizeNormalBar(bar, bar.targetValue);
  else
        bar.active = false;
        bar:Hide();
    end
end