-- ========================================================
--                           API
-- ========================================================

OverTimePredictor_Version = 2;

-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-- Use powerType < 0 to get health prediction
-- Returns total, dot, hot if unit is valid
-- Returns 0, 0, 0 if unit is invalid
-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function OverTimePredictor_GetBalance(powerType, unit)
	local guid = UnitGUID(unit);
	if guid then
		local hot, dot = 0, 0;
		for i=1, OverTimePredictor_GetNumEffects(guid) do
			local effect = OverTimePredictor_GetEffect(guid, i);
			if effect.powerType == powerType then
				local value = effect.total;
				if value < 0 then
					hot = hot - value;
				else
					dot = dot + value;
				end
			end
		end
		return hot - dot, hot, dot;
	end
	return 0, 0, 0;
end

-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-- Returns total, absorb, necro if unit is valid
-- Returns 0, 0, 0 if unit is invalid
-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function OverTimePredictor_GetAbsorb(unit)
	local guid = UnitGUID(unit);
	if guid then
		local absorb, necro = 0, 0;
		for i=1, OverTimePredictor_GetNumEffects(guid) do
			local effect = OverTimePredictor_GetEffect(guid, i);
			local value = effect.absorb;
			if value < 0 then
				necro = necro - value;
			else
				absorb = absorb + value;
			end
		end
		return absorb - necro, absorb, necro;
	end
	return 0, 0, 0;
end

-- ========================================================
--                         Functions
-- ========================================================

-- Maximal duration of the effect in order to be counted in DoT/HoT calculations
local MAX_DURATION = 61;

-- Set 2.06 if you want to assume all players take the crit meta
local CRITICAL_MULTIPLIER = 2.00; 

-- Sanity threshold to be sure that HoTs/DoTs with an initial instant tick are still handled correctly
local MIN_FIRST_TICK_DELAY = 0.10;

-- Used to avoid underestimating the number of ticks due to timestamp uncertainty and rounding errors
local CORRECTION_FACTOR = 1.05;

local UNITS = {
	"player",
	"pet",
	"pettarget",
	"target",
	"targettarget",
	"focus",
	"focustarget",
	"mouseover",
	"mouseovertarget",
};

do
	for i=1, 5 do
		tinsert(UNITS, "boss"..i);
		tinsert(UNITS, "arena"..i);
		tinsert(UNITS, "party"..i);
		tinsert(UNITS, "partypet"..i);
	end
	for i=1, 40 do
		tinsert(UNITS, "raid"..i);
		tinsert(UNITS, "raidpet"..i);
	end
end

local available = { };
--local COUNT = 0;

function OverTimePredictor_CreateTable()
	local table;
	if #available > 0 then
		table = available[#available];
		tremove(available, #available);
	else
		--COUNT = COUNT + 1;
		--if math.fmod(COUNT, 50) == 0 then
		--	print("alloc: "..COUNT);
		--end
		table = { };
	end
	return table;
end

function OverTimePredictor_DeleteTable(table)
	wipe(table);
	tinsert(available, table);
end

local function searchEffect(unit, filter, spellId)
	for i=1, 40 do
		local _, _, _, count, _, duration, expiration, _, _, _, spell = UnitAura(unit, i, filter);
		if not duration then
			break;
		elseif spell == spellId then
			return duration, expiration, count or 1;
		end
	end
	return nil, nil, 0;
end

local function searchEffectAll(list, unit, filter, spellId)
	local found = 0;
	for i=1, 40 do
		local _, _, _, count, _, duration, expiration, _, _, _, spell = UnitAura(unit, i, filter);
		if not duration then
			break;
		elseif spell == spellId then
			local entry = OverTimePredictor_CreateTable();
			tinsert(entry, duration);
			tinsert(entry, expiration);
			tinsert(entry, count or 1);
			tinsert(list, entry);
			found = found + 1;
		end
	end
	return found;
end

function OverTimePredictor_SearchEffect(unit, spellId)
	local duration, expiration, count = searchEffect(unit, "HELPFUL", spellId);
	if not duration then duration, expiration, count = searchEffect(unit, "HARMFUL", spellId); end
	return duration, expiration, count;
end

function OverTimePredictor_SearchEffectAll(list, unit, spellId)
	if searchEffectAll(list, unit, "HELPFUL", spellId) == 0 then
		searchEffectAll(list, unit, "HARMFUL", spellId);
	end
end

function OverTimePredictor_HandleAddEffect(timestamp, sourceGUID, destGUID, spellId, dose, absorbed)
	spellId = OverTimePredictor_GetEffectiveSpellId(spellId, sourceGUID, destGUID);

	local info = OverTimePredictor_GetSpellInfo(spellId);

	local effect = OverTimePredictor_CreateTable();

	effect.timestamp = timestamp;
	effect.source = sourceGUID;
	effect.spell = spellId;
	effect.duration = 0;
	effect.count = dose;
	effect.absorb = 0;
	effect.value = 0;
	effect.interval = 0;
	effect.school = 0;
	effect.powerType = -1;
	effect.tick = 0;
	effect.total = 0;

	if absorbed then effect.absorb = absorbed * info.absorbType; end

	OverTimePredictor_AddEffect(destGUID, effect);

	OverTimePredictor_ScanDuration(destGUID, effect);

	OverTimePredictor_RecomputeTotal(effect);
end

function OverTimePredictor_HandleRefreshEffect(timestamp, sourceGUID, destGUID, spellId, dose, absorbed)
	spellId = OverTimePredictor_GetEffectiveSpellId(spellId, sourceGUID, destGUID);

	local effect = OverTimePredictor_GetEffect(destGUID, sourceGUID, spellId);
	if effect then
		local info = OverTimePredictor_GetSpellInfo(spellId);

		--effect.timestamp = timestamp;
		effect.count = dose;
		effect.tick = 0;

		if absorbed then effect.absorb = absorbed * info.absorbType; end

		OverTimePredictor_ScanDuration(destGUID, effect);

		OverTimePredictor_RecomputeTotal(effect);
	end

	--print("refresh effect", timestamp, sourceGUID, destGUID, spellId, dose, absorbed);
end

function OverTimePredictor_HandleRemoveEffect(timestamp, sourceGUID, destGUID, spellId, dose, absorbed)
	spellId = OverTimePredictor_GetEffectiveSpellId(spellId, sourceGUID, destGUID);

	local _, index = OverTimePredictor_GetEffect(destGUID, sourceGUID, spellId);
	if index > 0 then
		OverTimePredictor_RemoveEffect(destGUID, index);
	end
end

function OverTimePredictor_HandleEffectTick(timestamp, sourceGUID, destGUID, spellId, school, powerType, amount, critical)
	-- powerType < 0 -> health
	-- critical = false/true/nil (nil if unknown)

	spellId = OverTimePredictor_GetEffectiveSpellId(spellId, sourceGUID, destGUID);

	local effect = OverTimePredictor_GetEffect(destGUID, sourceGUID, spellId);
	if effect then
		local interval = abs(timestamp - effect.timestamp);
		if interval > MIN_FIRST_TICK_DELAY then
			if effect.interval == 0 then
				effect.interval = timestamp - effect.timestamp;
			end

			effect.tick = effect.tick + 1;
		end

		if amount ~= 0 then
			if critical then
				amount = amount / CRITICAL_MULTIPLIER;
			end

			effect.value = amount;
		end

		effect.school = school;
		effect.powerType = powerType;

		OverTimePredictor_ScanDuration(destGUID, effect);

		OverTimePredictor_RecomputeTotal(effect);
	end

	--print("tick effect", timestamp, sourceGUID, destGUID, spellId, powerType, amount, critical);
end

function OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, spellId, absorbed)
	for i=1, OverTimePredictor_GetNumEffects(destGUID) do
		local effect = OverTimePredictor_GetEffect(destGUID, i);
		if effect then
			local mod = 0;
			if absorbed > 0 and effect.absorb > 0 then
				mod = min(absorbed, effect.absorb);
			elseif absorbed < 0 and effect.absorb < 0 then
				mod = max(absorbed, effect.absorb);
			end
			if mod ~= 0 then
				effect.absorb = effect.absorb - mod;
				absorbed = absorbed - mod;
				if absorbed == 0 then break; end
			end
		end
	end

	--print("absorb event", timestamp, sourceGUID, destGUID, spellId, absorbed);
end

function OverTimePredictor_RecomputeTotal(effect)
	local total = 0;
	if effect.interval > 0 and effect.duration < MAX_DURATION then
		local numTicks = floor(effect.duration / effect.interval * CORRECTION_FACTOR);
		local remainingTicks = max(0, numTicks - effect.tick);
		local critRate = OverTimePredictor_GetCritRate(effect.spell, effect.source, effect.school, effect.powerType);

		total = remainingTicks * effect.value * (1 - critRate + CRITICAL_MULTIPLIER * critRate);

		--print(effect.spell, effect.duration, effect.interval, numTicks, remainingTicks, critRate, total);
	end
	effect.total = total;
end

local function scanDuration(unit, effect)
	local best = 0;
	local delta = math.huge;
	local list = OverTimePredictor_CreateTable();

	OverTimePredictor_SearchEffectAll(list, unit, effect.spell);

	for i=#list, 1, -1 do
		local start = list[i][2] - list[i][1];
		local candidate_delta = abs(start - effect.timestamp);

		if candidate_delta < delta then
			best = list[i][1];
			delta = candidate_delta;
		end

		OverTimePredictor_DeleteTable(list[i]);
	end

	OverTimePredictor_DeleteTable(list);

	if best > 0 then
		if effect.duration == 0 then
			local info = OverTimePredictor_GetSpellInfo(effect.spell);

			if info.onDurationFound then
				info.onDurationFound(effect, unit);
			end
		end

		effect.duration = best;

		--print("found "..effect.duration.." for spell "..effect.spell);
	end
end

function OverTimePredictor_ScanDuration(destGUID, effect)
	for i=1, #UNITS do
		if UnitGUID(UNITS[i]) == destGUID then
			scanDuration(UNITS[i], effect);
			break;
		end
	end
end

function OverTimePredictor_Prune(unit)
	local guid = UnitGUID(unit);
	if guid then
		local count = OverTimePredictor_GetNumEffects(guid);
		for i=count, 1, -1 do
			local effect = OverTimePredictor_GetEffect(guid, i);
			local duration = OverTimePredictor_SearchEffect(unit, effect.spell);
			if not duration then
				OverTimePredictor_RemoveEffect(guid, i);
			else
				scanDuration(unit, effect);

				OverTimePredictor_RecomputeTotal(effect);
			end
		end
	end
end

-- ========================================================
--                         Handlers
-- ========================================================

function OverTimePredictor_OnCombatLogEvent(timestamp, event, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, ...)
	if not sourceGUID or not destGUID then return; end

	local spellId, spellName, spellSchool;
	local amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, overhealing;
	local missType, isOffHand, amountMissed;
	local extraAmount, remainingPoints;
	local powerType, alternatePowerType, _;

	if ( strsub(event, 1, 5) == "SPELL" ) then
		spellId, spellName, spellSchool = ...;

		if ( event == "SPELL_MISSED" ) then
			missType,  isOffHand, multistrike, amountMissed = select(4, ...);
			if missType == "ABSORB" then 
				OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, spellId, amountMissed or 0);
			end
			
		elseif ( event == "SPELL_DAMAGE" ) then
			amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand, multistrike = select(4, ...);
			OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, spellId, absorbed or 0);
			
		elseif ( event == "SPELL_HEAL" ) then
			amount, overhealing, absorbed, critical, multistrike = select(4, ...);
			OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, spellId, -(absorbed or 0));
			
		elseif ( event == "SPELL_PERIODIC_MISSED" ) then
			missType, isOffHand, multistrike, amountMissed = select(4, ...);
			OverTimePredictor_HandleEffectTick(timestamp, sourceGUID, destGUID, spellId, spellSchool, -1, amountMissed or 0, nil);
			if missType == "ABSORB" then 
				OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, spellId, amountMissed or 0);
			end

		elseif ( event == "SPELL_PERIODIC_DAMAGE" ) then
			amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand, multistrike = select(4, ...);
			OverTimePredictor_HandleEffectTick(timestamp, sourceGUID, destGUID, spellId, school, -1, amount, critical);
			OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, spellId, absorbed or 0);

		elseif ( event == "SPELL_PERIODIC_HEAL" ) then
			amount, overhealing, absorbed, critical, multistrike = select(4, ...);
			OverTimePredictor_HandleEffectTick(timestamp, sourceGUID, destGUID, spellId, spellSchool, -1, -amount, critical);
			OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, spellId, -(absorbed or 0));

		elseif ( event == "SPELL_PERIODIC_DRAIN" ) then
			amount, powerType, extraAmount, alternatePowerType = select(4, ...);
			OverTimePredictor_HandleEffectTick(timestamp, sourceGUID, destGUID, spellId, spellSchool, powerType, amount, critical);

		elseif ( event == "SPELL_PERIODIC_LEECH" ) then
			amount, powerType, extraAmount, alternatePowerType = select(4, ...);
			OverTimePredictor_HandleEffectTick(timestamp, sourceGUID, destGUID, spellId, spellSchool, powerType, amount, nil);

		elseif ( event == "SPELL_PERIODIC_ENERGIZE" ) then 
			amount, powerType, alternatePowerType = select(4, ...);
			OverTimePredictor_HandleEffectTick(timestamp, sourceGUID, destGUID, spellId, spellSchool, powerType, -amount, nil);

		elseif ( event == "SPELL_AURA_APPLIED" ) then
			_, remainingPoints = select(4, ...);
			OverTimePredictor_HandleAddEffect(timestamp, sourceGUID, destGUID, spellId, 1, remainingPoints);

		elseif ( event == "SPELL_AURA_REFRESH" ) then
			_, remainingPoints = select(4, ...);
			OverTimePredictor_HandleRefreshEffect(timestamp, sourceGUID, destGUID, spellId, 1, remainingPoints);

		elseif ( event == "SPELL_AURA_REMOVED" ) then
			_, remainingPoints = select(4, ...);
			OverTimePredictor_HandleRemoveEffect(timestamp, sourceGUID, destGUID, spellId, 0, remainingPoints);
		end

	elseif ( event == "SWING_MISSED" ) then 
		missType, isOffHand, multistrike, amountMissed = ...;
		if missType == "ABSORB" then 
			OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, 0, amountMissed or 0);
		end

	elseif ( event == "SWING_DAMAGE" ) then 
		amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand, multistrike = ...;
		OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, 0, absorbed or 0);

	elseif ( event == "RANGE_MISSED" ) then 
		missType, isOffHand, multistrike, amountMissed = select(4, ...);
		if missType == "ABSORB" then 
			OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, 0, amountMissed or 0);
		end

	elseif ( event == "RANGE_DAMAGE" ) then 
		amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand, multistrike = select(4, ...);
		OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, 0, absorbed or 0);

	elseif ( event == "DAMAGE_SHIELD_MISSED" ) then
		spellId, spellName, spellSchool, missType, isOffHand, multistrike, amountMissed = ...; -- FIXME: might be unsupported
		if missType == "ABSORB" then 
			OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, spellId, amountMissed or 0);
		end

	elseif ( event == "DAMAGE_SHIELD" ) then
		spellId, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed, critical = ...;
		OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, spellId, absorbed or 0);

	elseif ( event == "DAMAGE_SPLIT" ) then
		spellId, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed, critical = ...;
		OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, spellId, absorbed or 0);

	elseif ( event == "ENVIRONMENTAL_DAMAGE" ) then
		_, amount, overkill, school, resisted, blocked, absorbed, critical = ...;
		OverTimePredictor_HandleAbsorb(timestamp, sourceGUID, destGUID, 0, absorbed or 0);
	end
end

function OverTimePredictor_OnLoad(self)
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	self:RegisterEvent("UNIT_AURA");
	self:RegisterEvent("UNIT_TARGET");
	self:RegisterEvent("PLAYER_TARGET_CHANGED");
	self:RegisterEvent("PLAYER_FOCUS_CHANGED");
	self:RegisterEvent("PLAYER_ENTERING_WORLD");
end

function OverTimePredictor_OnEvent(self, event, ...)
	if event == "COMBAT_LOG_EVENT_UNFILTERED" then
		OverTimePredictor_OnCombatLogEvent(...);

	elseif event == "UNIT_AURA" then
		local unit = ...;
		OverTimePredictor_Prune(unit);

	elseif event == "UNIT_TARGET" then
		local unit = ...;
		OverTimePredictor_Prune(unit.."target");

	elseif event == "PLAYER_TARGET_CHANGED" then
        	OverTimePredictor_Prune("target");
        	OverTimePredictor_Prune("targettarget");

	elseif event == "PLAYER_FOCUS_CHANGED" then
        	OverTimePredictor_Prune("focus");
        	OverTimePredictor_Prune("focustarget");

	elseif event == "PLAYER_ENTERING_WORLD" then
		OverTimePredictor_WipeDatabase();
	end
end